/* Name: main.c
 * Project: Presentation remote - receiver
 * Author: jiristepanovsky.cz
 * Creation Date: 2013-12-1
 */

#include "fcpu.h"

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>

#include "usbdrv.h"
#include "oddebug.h"

#define RX_STATE (PINB&0b00000010)	//Receiver input state

static uchar reportBuffer[2];	//Buffer for HID reports
static uchar idleRate;	//In 4 ms units

uchar pressedKey = 0;
uchar dataAvailable = 0;

/* ------------------------------------------------------------------------- */
const unsigned char rf_timing_id[25] PROGMEM = { 0, 180, 42, 42, 110, 42, 42, 42, 42,
                                                    110, 42, 42, 110, 110,
                                                    42, 42, 110, 110, 42, 42,
                                                    42, 42, 110, 110, 110 };
const unsigned char rf_timing_key[4][11] PROGMEM ={{110, 110, 42, 42,
                                                    110, 42, 42, 110, 42, 200, 200}, // Middle
                                                   {110, 110, 110,
                                                    110, 42, 42, 42, 42, 42, 200, 200}, // left
                                                   {110, 110, 42, 42,
                                                    42, 42, 42, 42, 42, 42, 42}, // right
                                                    {42, 42, 42, 42, 110,
                                                    110, 42, 42, 42, 42, 42}}; // bottom right
/* ------------------------------------------------------------------------- */

const PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {
/* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

static void buildReport(void)
{
	uchar key = 0;

	if(dataAvailable > 1){
        // Process received key from 433 Mhz receiver module
		switch(pressedKey){
            case 0: key = 0x27; break; //0
            case 1: key = 0x1E; break; //1
            case 2: key = 0x1F; break; //2
            case 3: key = 0x20; break; //3
            case 4: key = 0x21; break; //4
            case 5: key = 0x22; break; //5
            case 6: key = 0x23; break; //6
            case 7: key = 0x24; break; //7
            case 8: key = 0x25; break; //8
            case 9: key = 0x26; break; //9
            case 12: key = 0x05; break; //B
            case 13: key = 0x50; break; //LEFT
            case 14: key = 0x4F; break; //RIGHT
            case 15: key = 0x3E; break; //F5
            default: key = 0x2D; break;
		}
	} else {
        // No key pressed
		pressedKey = 0;
	}

	dataAvailable--;
    reportBuffer[0] = 0;
    reportBuffer[1] = key;
}

/* -------------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* -------------------------------------------------------------------------------- */

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport();
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

void    usbEventResetReady(void)
{
}

int main(void)
{

	uchar i;
	uchar time, _time;
	uchar rf_state = 0;
	uchar rf_button = 0, rf_timeout = 0;
    uchar rf_input = 0;
    uchar keys, key[4]; // keys

    uchar debugl = 0;
    unsigned short debug = 0;

	//odDebugInit();
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    wdt_enable(WDTO_1S);
    sei();
    
    // Timer init
    TCCR0A = 0;
    TCCR0B = (1<<CS02) | (1<<CS00); // Prescaller 1024
    TCCR1 = (1<<CS12) | (1<<CS11);  //Prescaller 32
    TCNT0 = 0;
	TCNT1 = 0;
	//TIFR |= (1 << TOV1); // clear overflow
    
    // Main loop

    while(1){
        wdt_reset();
        usbPoll();
		
        // Emulate key press
		if(usbInterruptIsReady() && dataAvailable){
        	buildReport();
           	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        }
        
        // RF receiver, overflow after ~21ms
        TCNT0 = 0;
        TCNT1 = 0;
        TIFR |= (1 << TOV0) | (1 << TOV1); // clear flags
        rf_state = 0;
        while(!(TIFR&(1<<TOV0))){
            if(rf_input != RX_STATE){
                // pin change
                rf_input = RX_STATE;
                if(TIFR&(1<<TOV1)){
                    TIFR |= (1 << TOV1); // clear overflow
                    rf_state = 0;
                } else {
                    // get time
                    time = TCNT1;
                    TCNT1 = 0;
                    if(!rf_state){
                        if(RX_STATE)
                            rf_state = 1;
                    } else {
                        if(rf_state < 25){
                            // id
                            _time = pgm_read_byte(&(rf_timing_id[rf_state]));
                            if(time > (_time-30) && time < (_time+30)){
                                rf_state++;
                                if(rf_state >= 25){
                                    for(i = 0; i < 4; i++)
                                        key[i] = 1;
                                }
                            } else {
                                //if(!debugl && rf_state > 2){
                                //    debug = (1000*rf_state)+time;
                                //    debugl = 6;
                                //}
                                rf_state = 1;
                            }
                        } else {
                            // key
                            keys = 0;
                            for(i = 0; i < 4; i++){
                                if(key[i]){
                                    _time = pgm_read_byte(&(rf_timing_key[i][rf_state-25]));
                                    if(time < (_time-30) || time > (_time+30)){
                                        key[i] = 0;
                                    }
                                }
                                keys += key[i];
                            }
                            if(keys){
                                rf_state++;
                                if((rf_state >= 34 && (key[0] || key[1])) || rf_state >= 36){
                                    // button pressed
                                    if(key[0]){
                                        pressedKey = 12; // Key 1
                                    }
                                    if(key[1]){
                                        pressedKey = 13; // Key 2
                                    }
                                    if(key[2]){
                                        pressedKey = 14; // Key 3
                                    }
                                    if(key[3]){
                                        pressedKey = 15; // Key 4
                                    }
                                    //if(!debugl){
                                    //    debug = 0;
                                    //    debugl = 2;
                                    //}
                                    if(!rf_button || rf_button == 2)
                                        rf_button++;
                                    rf_state = 0;
                                }
                            } else {
                                //if(!debugl){
                                //    debug = (1000*rf_state)+time;
                                //    debugl = 6;
                                //}
                                rf_state = 1;
                            }
                        }
                    }
                }
            }
        }
        
        /*if(rf_timeout < 40)
            rf_timeout++;
        if(debugl && !dataAvailable && rf_timeout > 15){
            rf_timeout = 0;
            dataAvailable = 2;
            debugl--;
            if(debugl == 5){
                pressedKey = (debug/10000)%10;
            } else if(debugl == 4){
                pressedKey = (debug/1000)%10;
            } else if(debugl == 3){
                pressedKey = (debug/100)%10;
            } else if(debugl == 2){
                pressedKey = (debug/10)%10;
            } else if(debugl == 1){
                pressedKey = debug%10;
            } else {
                pressedKey = 99;
            }
        }*/
        
        if(rf_button == 2){
            rf_timeout++;
            if(rf_timeout > 4)
                rf_button = 0;
        }
        if(rf_button == 3){
            rf_button = 2;
            rf_timeout = 0;
        }
        if(rf_button == 1){
            if(!dataAvailable)
                dataAvailable = 2;
            rf_button = 2;
            rf_timeout = 0;
        }
        
	}
   	return 0;
}


/*******************************************************************
Keyboard codes
	0x00	Reserved (no event indicated)
	0x01	Keyboard ErrorRollOver
	0x02	Keyboard POSTFail
	0x03	Keyboard ErrorUndefined
	0x04	Keyboard a and A
	0x05	Keyboard b and B
	0x06	Keyboard c and C
	0x07	Keyboard d and D
	0x08	Keyboard e and E
	0x09	Keyboard f and F
	0x0A	Keyboard g and G
	0x0B	Keyboard h and H
	0x0C	Keyboard i and I
	0x0D	Keyboard j and J
	0x0E	Keyboard k and K
	0x0F	Keyboard l and L
	0x10	Keyboard m and M
	0x11	Keyboard n and N
	0x12	Keyboard o and O
	0x13	Keyboard p and P
	0x14	Keyboard q and Q
	0x15	Keyboard r and R
	0x16	Keyboard s and S
	0x17	Keyboard t and T
	0x18	Keyboard u and U
	0x19	Keyboard v and V
	0x1A	Keyboard w and W
	0x1B	Keyboard x and X
	0x1C	Keyboard y and Y
	0x1D	Keyboard z and Z
	0x1E	Keyboard 1 and !
	0x1F	Keyboard 2 and @
	0x20	Keyboard 3 and #
	0x21	Keyboard 4 and $
	0x22	Keyboard 5 and %
	0x23	Keyboard 6 and ^
	0x24	Keyboard 7 and &
	0x25	Keyboard 8 and *
	0x26	Keyboard 9 and (
	0x27	Keyboard 0 and )
	0x28	Keyboard Return (ENTER)
	0x29	Keyboard ESCAPE
	0x2A	Keyboard DELETE (Backspace)
	0x2B	Keyboard Tab
	0x2C	Keyboard Spacebar
	0x2D	Keyboard - and (underscore)
	0x2E	Keyboard = and +
	0x2F	Keyboard [ and {
	0x30	Keyboard ] and }
	0x31	Keyboard \ and |
	0x32	Keyboard Non-US # and ~
	0x33	Keyboard ; and :
	0x34	Keyboard ' and "
	0x35	Keyboard Grave Accent and Tilde
	0x36	Keyboard, and <
	0x37	Keyboard . and >
	0x38	Keyboard / and ?
	0x39	Keyboard Caps Lock
	0x3A	Keyboard F1
	0x3B	Keyboard F2
	0x3C	Keyboard F3
	0x3D	Keyboard F4
	0x3E	Keyboard F5
	0x3F	Keyboard F6
	0x40	Keyboard F7
	0x41	Keyboard F8
	0x42	Keyboard F9
	0x43	Keyboard F10
	0x44	Keyboard F11
	0x45	Keyboard F12
	0x46	Keyboard PrintScreen
	0x47	Keyboard Scroll Lock
	0x48	Keyboard Pause
	0x49	Keyboard Insert
	0x4A	Keyboard Home
	0x4B	Keyboard PageUp
	0x4C	Keyboard Delete Forward
	0x4D	Keyboard End
	0x4E	Keyboard PageDown
	0x4F	Keyboard RightArrow
	0x50	Keyboard LeftArrow
	0x51	Keyboard DownArrow
	0x52	Keyboard UpArrow
	0x53	Keypad Num Lock and Clear
	0x54	Keypad /
	0x55	Keypad *
	0x56	Keypad -
	0x57	Keypad +
	0x58	Keypad ENTER
	0x59	Keypad 1 and End
	0x5A	Keypad 2 and Down Arrow
	0x5B	Keypad 3 and PageDn
	0x5C	Keypad 4 and Left Arrow
	0x5D	Keypad 5
	0x5E	Keypad 6 and Right Arrow
	0x5F	Keypad 7 and Home
	0x60	Keypad 8 and Up Arrow
	0x61	Keypad 9 and PageUp
	0x62	Keypad 0 and Insert
	0x63	Keypad . and Delete
	0x64	Keyboard Non-US \ and |
	0x65	Keyboard Application
	0x66	Keyboard Power
	0x67	Keypad =
	0x68	Keyboard F13
	0x69	Keyboard F14
	0x6A	Keyboard F15
	0x6B	Keyboard F16
	0x6C	Keyboard F17
	0x6D	Keyboard F18
	0x6E	Keyboard F19
	0x6F	Keyboard F20
	0x70	Keyboard F21
	0x71	Keyboard F22
	0x72	Keyboard F23
	0x73	Keyboard F24
	0x74	Keyboard Execute
	0x75	Keyboard Help
	0x76	Keyboard Menu
	0x77	Keyboard Select
	0x78	Keyboard Stop
	0x79	Keyboard Again
	0x7A	Keyboard Undo
	0x7B	Keyboard Cut
	0x7C	Keyboard Copy
	0x7D	Keyboard Paste
	0x7E	Keyboard Find
	0x7F	Keyboard Mute
	0x80	Keyboard Volume Up
	0x81	Keyboard Volume Down
	0x82	Keyboard Locking Caps Lock
	0x83	Keyboard Locking Num Lock
	0x84	Keyboard Locking Scroll Lock
	0x85	Keypad Comma
	0x86	Keypad Equal Sign
	0x87	Keyboard International1
	0x88	Keyboard International2
	0x89	Keyboard International3
	0x8A	Keyboard International4
	0x8B	Keyboard International5
	0x8C	Keyboard International6
	0x8D	Keyboard International7
	0x8E	Keyboard International8
	0x8F	Keyboard International9
	0x90	Keyboard LANG1
	0x91	Keyboard LANG2
	0x92	Keyboard LANG3
	0x93	Keyboard LANG4
	0x94	Keyboard LANG5
	0x95	Keyboard LANG6
	0x96	Keyboard LANG7
	0x97	Keyboard LANG8
	0x98	Keyboard LANG9
	0x99	Keyboard Alternate Erase
	0x9A	Keyboard SysReq/Attention
	0x9B	Keyboard Cancel
	0x9C	Keyboard Clear
	0x9D	Keyboard Prior
	0x9E	Keyboard Return
	0x9F	Keyboard Separator
	0xA0	Keyboard Out
	0xA1	Keyboard Oper
	0xA2	Keyboard Clear/Again
	0xA3	Keyboard CrSel/Props
	0xA4	Keyboard ExSel
	0xE0	Keyboard LeftControl
	0xE1	Keyboard LeftShift
	0xE2	Keyboard LeftAlt
	0xE3	Keyboard Left GUI
	0xE4	Keyboard RightControl
	0xE5	Keyboard RightShift
	0xE6	Keyboard RightAlt
	0xE7	Keyboard Right GUI
*******************************************************************/
